C====================================================================
C DETERMINE G(mu_t) IN (B.1), USING THE PROJECTION METHOD
C====================================================================
      USE NUMERICAL_LIBRARIES
      IMPLICIT NONE
      INTEGER NQUD,IW,NFIX,I,NR
      PARAMETER (NQUD=10,IW=4,NFIX=0,NR=100000)
      REAL*8 DEFAULTA,DEFAULTB,QXFIX(0:NFIX),PR,EC,ED,ESigD,GD,
     &       LAM,CONS(NR),PD(NR),EXRET(NR),COR(10)
C      REAL*8 DRNUNF,DRNNOF
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

C--COMUPTE THE GAUSSIAN QUADRATURE POINTS AND WEIGHTS

      CALL ERSET(3,0,-1)                   ! TURNING OFF WARNING MESSAGES

      DEFAULTA= 2.0D0
      DEFAULTB = 2.0D0
      CALL DGQRUL(NQUD,IW,DEFAULTA,DEFAULTA,NFIX,QXFIX,QX,QW)    ! QX CONTAINS POINTS, QW CONTAINS WEIGHTS

C--PARAMETERS OF DATA (CONSUMPTION) GENERATING PROCESS FROM cECCHETTI ET AL (2000, AER)

      Q = 0.516D0           ! BAD-TO-BAD STATE (STATE 1)
      P = 0.978D0           ! GOOD-TO-GOOD STATE (STATE 2)
      KaC(1) = -0.06785D0   ! BAD STATE CONS. GROWTH RATE
      KaC(2) = 0.02251D0    ! GOOD STATE CONS. GROWTH RATE
      SigC = 0.03127D0      ! STD OF CONS GROWTH RATE

      PR = (1-P)/(2-P-Q)    ! STATIONARY PROB OF BAD STATE = 1
      EC = PR*KaC(1) + (1-PR)*KaC(2)          ! EXPECTED CONS GROWTH RATE = 1.86%

C--PARAMETERS OF DATA DIVIDEND PROCESS

      LAM = 2.74D0          ! LEVERAGE EFFECT COEFFICIENT
      ESigD = 0.12D0        ! STD OF DIV CONDITION ON STATE, IMPLYING UNCOND STD = 0.13019 = 13%

      Gd = EC - LAM*EC           ! TO MAKE SURE E[DIV GROWTH RATE] = CONS GROWTH = EC = 1.86%.
      KaD(1) = LAM*KaC(1) + Gd   ! BAD STATE DIV. GROWTH RATE
      KaD(2) = LAM*KaC(2) + Gd   ! GOOD STATE DIV. GROWTH RATE
      SigD = LAM*SigC            ! STD OF THE LEVERED PART OF THE DIV PROCESS IN (14)
      SigU = DSQRT(ESigD**2-LAM**2*SigC**2)  ! THIS IS SigD IN (14) NAMED DIFFERENTLY = 0.084018

C--PREFERENCE MODEL PARAMETERS

      BET = 0.974929671636469D0
      GAM = 2.0D0
      RHO = 1.0D0/1.5D0
      ETA = 8.8642912575999D0

      IF (ETA.LT.GAM) STOP 'ETA HAS TO >= GAM'

C--RETURN THE DENSITY RATIOS FOR A GIVEN PAIR (CONS RATE, DIV RATE) USED IN BAYSIAN
C--UPDATING TO A COMMON BLOCK, USED REPEATEDLY

      CALL DEN_RATIO_GRID(NQUD,QX,FY21)

C--FULL MODEL WITH LEARNING AND AMBIGUITY

      CALL FULL_MODEL_SERIES(CONS,PD,EXRET)

      CALL CORRELATION(CONS,PD,EXRET,COR)

      WRITE(6,10) COR(1),COR(2),COR(3),COR(4),COR(5)
10    FORMAT(1X,5F7.2)

      STOP
      END
C=====================================================================================
C RETURN UNCONDITIONAL MOMENTS OF THE LEARNING AND AMBIGUITY MODEL, WHICH INCLUDES
C LEARNING BUT WITHOUT AMBIGUITY MODEL AS A SPECIAL CASE BY SETTING: ETA = GAMMA
C=====================================================================================
      SUBROUTINE FULL_MODEL_COEF(CN_G,CN_RM)
      IMPLICIT NONE
      INTEGER NR,NEQ,NORD,IP(6),I
      PARAMETER (NEQ=21)
      REAL*8 CN_G(0:100),CN_RM(0:100),G1,G2,VP1,VP2
      REAL*8 XG(NEQ),XS(NEQ),FS(NEQ),RP(5),X(NEQ),FV(NEQ)
      DATA XS/NEQ*1.0D0/,FS/NEQ*1.0D0/        ! using DEFAULT VALUES AND INITIALIZE CN TO 0
      EXTERNAL REQS_G,REQS_RM

      CALL FULL_INFO_COEF(G1,G2,VP1,VP2)   ! FULL INFO MODEL PROVIDE INITIAL GUESS

      XG(1) = G2             ! SET CN(0) TO FULL-INFO MODEL HIGH STATE VALUE
      DO I = 2, NEQ                ! SET OTHERS INITIALLY TO ZERO
       XG(I) = 0.0D0
      ENDDO

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(REQS_G,NEQ,XG,XS,FS,IP,RP,X,FV)   ! IMSL ROUTINE SOLVING nonlinear equations

C--FINAL CN(0:NORD) COEFFICIENTS

      NORD = NEQ - 1
      DO I = 0, NORD
       CN_G(I) = X(I+1)
      ENDDO

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR \varphi(mu_t).

      XG(1) = VP2            ! INITIAL GUESS OF PRICE/DIV RATIO, SET TO FULL-INFO MODEL HIGH STATE VALUE
      DO I = 2, NEQ                ! SET OTHERS INITIALLY TO ZERO
       XG(I) = 0.0D0
      ENDDO

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(REQS_RM,NEQ,XG,XS,FS,IP,RP,X,FV)  ! IMSL ROUTINE SOLVING nonlinear equations

C--FINAL CN_RM(0:NORD) COEFFICIENTS

      NORD = NEQ - 1
      DO I = 0, NORD
       CN_RM(I) = X(I+1)
      ENDDO

      RETURN
      END
C=====================================================================================
C RETURN UNCONDITIONAL MOMENTS OF THE FULL INFORMATION MODEL
C=====================================================================================
      SUBROUTINE FULL_INFO_COEF(G1,G2,VP1,VP2)
      IMPLICIT NONE
      INTEGER NEQ,IP(6)
      PARAMETER (NEQ=2)
      REAL*8 G1,G2,VP1,VP2
      REAL*8 GTC,G0
      REAL*8 XG(NEQ),XS(NEQ),FS(NEQ),RP(5),X(NEQ),FV(NEQ)
      DATA XS/NEQ*1.0D0/,FS/NEQ*1.0D0/          ! using DEFAULT VALUES AND INITIALIZE CN TO 0
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      EXTERNAL FULL_REQS_G

C--CHECK TRANSVERSALITY CONDITION. ASSUME G(Z_t) IS A CONSTANT AND SOLVE THE
C--OBJECTIVE FUNCTION (RECURSIVE) EQUATION FOR AN INITIAL GUESS. ASSUME STATE
C--AT t+1 IS 2 (HIGH GROWTH STATE).

      GTC = BET*DEXP((KaC(2)+0.5*(1-GAM)*SigC**2)*(1-RHO))
      IF (GTC.GE.1.0D0) THEN
       WRITE(6,*) 'GTC ',GTC
       WRITE(6,*) 'MAY VIOLATE TRANS. CONDITION. CHECK TO MAKE SURE'
       STOP
      ENDIF

      G0 = 1.0D0
      IF (GTC.LT.1.0D0) G0 = ((1-BET)/(1-GTC))**(1/(1-RHO))   ! IF GTC < 1, TRANSVERSALITY CONDITION OK

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR G(Z_t): G(1) AND G(2).

      XG(1) = G0                   ! SET INITIAL GUESS TO G0
      XG(2) = G0

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(FULL_REQS_G,NEQ,XG,XS,FS,IP,RP,X,FV)   ! IMSL ROUTINE SOLVING nonlinear equations
      G1 = X(1)
      G2 = X(2)

      IF (MIN(G1,G2).LT.0.0D0) THEN
       WRITE(6,*) 'G1, G2 ', G1,G2
       WRITE(6,*) 'G1, G2 HAS TO BE POTIVE'
       STOP
      ENDIF

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR \varphi(Z_t): \varphi(1) AND
C--\varphi(2). ASSUME G(Z_t) IS A CONSTANT AND SOLVE THE EULAER EQUATION FOR AN INITIAL GUESS.
C--ASSUME STATE AT t+1 IS 2 (HIGH GROWTH STATE).

      CALL FULL_REQS_RM(G1,G2,VP1,VP2)            ! solve P/D ratio (VP1, VP2)

      RETURN
      END
C================================================================================================
C RETURN THE TWO FACTORS IN COMPUTING RISKFREE RATE AT EACH TIME POINT, GIVEN THE STATE PROB DIST.
C================================================================================================
      FUNCTION RISKFREE(NORD,MUt,CN_G)
      IMPLICIT NONE
      INTEGER NORD,NQUD,IC,ID,IS,ISTATE
      PARAMETER (nqud=10)
      REAL*8 RISKFREE,MUt,CN_G(0:100)
      REAL*8 Zc,QWC,PKERNEL,KERNEL,HS(2),FUNC_G,PI,EM1,EM2
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW

      ISTATE = 1
      HS(1) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 1
      ISTATE = 2
      HS(2) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 2

      EM1 = 0.0D0
      EM2 = 0.0D0
      DO IC = 1, NQUD
       Zc = QX(IC)
       QWC = QW(IC)
       IS = 1
       KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
       EM1 = EM1 + KERNEL*QWC       ! GUANSIAN 2D QUDRATURE, COMPUTE E[M]
       IS = 2
       KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
       EM2 = EM2 + KERNEL*QWC       ! GUANSIAN 2D QUDRATURE
      ENDDO

      PI = 2*DASIN(1.0D0)
      EM1 = EM1/DSQRT(PI)                ! DIVIDE PI, COMING FROM 2D STD NORMAL DENSITY FUNCTION
      EM2 = EM2/DSQRT(PI)
      RISKFREE = 1/(MUt*EM1+(1-MUt)*EM2)       ! RF = 1/E[M]

      RETURN
      END
C======================================================================
C RETURN THE PRICING KERNEL, GIVEN MUt,Zc,Zd, AND STATE IS = (1, OR 2),
C GROWTH RATES: Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc,
C Gd = KaD(IS) + SigD*DSQRT(2.0D0)*(COR*Zc+DSQRT(1-COR**2)*Zd).
C======================================================================
      FUNCTION PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
      IMPLICIT NONE
      INTEGER IS,IC,NQUD,NORD,ISTATE
      PARAMETER (nqud=10)
      REAL*8 PKERNEL,CN_G(0:100),MUt,HS(2),Zc
      REAL*8 Gc,FAC1,FAC2,FAC3,EX1,EX2,MUtp1,GMUtp1,CBSVN,
     &       Vtp1,RtVtp1,DR
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      Zc = QX(IC)

C--FIRST FACTOR: BET*Gc^{-RHO}

      Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc        ! LOG(C_{t+1}/C_t), CONS GROWTH RATE
      FAC1 = BET*DEXP(-RHO*Gc)

C--SECOND FACTOR: (Vtp1/RtVtp1)^(RHO-GAM).
C--HS(1:2) ARE THE TWO CONDITIONAL EXPECTATION IN Rt(V_{t+1}) IN (8). THE SAME
C--HS(1:2) (CONDITIONAL EXPECTATIONS) ARE USED IN (B.1) FOR THE VALUE FUNC.

      DR = FY21(IS,IC)                     !   DENSITY RATIO
      MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
      GMUtp1 = CBSVN(NORD,CN_G,2*MUtp1-1)       !   G(mu_{t+1})
      Vtp1 = GMUtp1*DEXP(Gc)                  !   V_{t+1}/C_t

      EX1 = (1-ETA)/(1-GAM)
      EX2 = 1/(1-ETA)
      RtVtp1 = (MUt*HS(1)**EX1+(1-MUt)*HS(2)**EX1)**EX2       ! Rt(v_{t+1})

      EX1 = RHO - GAM
      FAC2 = (Vtp1/RtVtp1)**EX1

C--THIRD FACTOR: (E_{t,J}[V_{t+1}^{1-GAM}])^{-(ETA-GAM)/(1-GAM)}/RtVtp1^{-(ETA-GAM)}

      EX1 = -(ETA-GAM)/(1-GAM)
      EX2 = -(ETA-GAM)
      FAC3 = HS(IS)**EX1/RtVtp1**EX2

C--CONDITIONAL PRICING KERNEL

      PKERNEL = FAC1*FAC2*FAC3

      RETURN
      END
C======================================================================
C RETURN NEQ RESIDUAL VALUES FOR THE SYSTEM OF NEQ EQUATIONS.
C
C INPUTS: NEQ---NUMBER OF RESIDUAL EQUATIONS = NUMBER OF COEF'S;
C         SN(NEQ)---COEF'S;
C         OTHER INPUT VALUES USED---PASSED THROUGH THE COMMON BLOCKS;
C
C OUTPUTS: RESIDUL(NEQ)---NEQ RESIDUALS, EACH FOR ONE PIt VALUE = NODAL VALUE OF CHEBYSHEV(NEQ)
C
C FUNCTIONS USED: CBSVN(N,CN,Z)---RETURN THE VALUE OF THE LAST TERM IN RESIDUAL R(PIt;CN);
C
C SUBROUTINE CALLED: DQDAGS---IMSL LIBRARY ROUTINE FOR INTEGRATION OF ONE-VARIABLE FUNCTION;
C
C======================================================================
      SUBROUTINE REQS_RM(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ,NORD,IS,I
      REAL*8 SN(NEQ),RC(NEQ)
      REAL*8 H1,H2,FUNC_RM,PI,MUt,RHSB2
      REAL*8 CN_G(0:100),CN_RM(0:100)
      COMMON /G_RM_COEF/CN_G,CN_RM

      NORD = NEQ - 1
      DO I = 0, NORD             ! PASS CN TO COMMON BLOCK, USED BY OTHER SUBROUTINES
       CN_RM(I) = SN(I+1)
      ENDDO

      PI = 2*DASIN(1.0D0)
      DO I = 1, NEQ               ! NEQ EQUATIONS TO DETERMINE C0,C1,....C(NEQ-1)
       MUt = 0.5 + 0.5*DCOS((NEQ-I+0.5)*PI/NEQ)    ! NEQ NODES OVER [0,1] OF CHEBYSHEV(NEQ)

C--RETURN THE INNER EXPECTATION IN (B.1) IS = 1, 2

       IS = 1
       H1 = FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)    ! STTAE IS = 1
       IS = 2
       H2 = FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)    ! STTAE IS = 2
       RHSB2 = MUt*H1 + (1-MUt)*H2                   !  RHS OF (B.2)
       RC(I) = RHSB2 - 1                   ! THE EULER EQUATION (B.2)
      ENDDO

      RETURN
      END
C========================================================================================
C RETURN NEQ RESIDUAL VALUES FOR THE SYSTEM OF NEQ EQUATIONS.
C
C INPUTS: NEQ---NUMBER OF RESIDUAL EQUATIONS = NUMBER OF COEF'S;
C         SN(NEQ)---COEF'S;
C         OTHER INPUT VALUES USED---PASSED THROUGH THE COMMON BLOCKS;
C
C OUTPUTS: RESIDUL(NEQ)---NEQ RESIDUALS, EACH FOR ONE PIt VALUE = NODAL VALUE OF CHEBYSHEV(NEQ)
C
C FUNCTIONS USED: CBSVN(N,CN,Z)---RETURN THE VALUE OF THE LAST TERM IN RESIDUAL R(PIt;CN);
C
C SUBROUTINE CALLED: DQDAGS---IMSL LIBRARY ROUTINE FOR INTEGRATION OF ONE-VARIABLE FUNCTION;
C
C======================================================================
      SUBROUTINE REQS_G(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ,NQUD,NORD,IS,I
      PARAMETER (nqud=10)
      REAL*8 SN(NEQ),RC(NEQ),CBSVN
      REAL*8 H1,H2,H21,FUNC_G,PI,MUt,CN_G(0:100),EX1,EX2,EX3,EX4,RHSB1
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      NORD = NEQ - 1
      DO I = 0, NORD             ! PASS CN TO COMMON BLOCK, USED BY OTHER SUBROUTINES
       CN_G(I) = SN(I+1)
      ENDDO

      PI = 2*DASIN(1.0D0)
      DO I = 1, NEQ               ! NEQ EQUATIONS TO DETERMINE C0,C1,....C(NEQ-1)
       MUt = 0.5 + 0.5*DCOS((NEQ-I+0.5)*PI/NEQ)    ! NEQ NODES OVER [0,1] OF CHEBYSHEV(NEQ)

C--RETURN THE INNER EXPECTATION IN (B.1) IS = 1, 2

       IS = 1
       H1 = FUNC_G(IS,MUt,NORD,CN_G)    ! STTAE IS = 1
       IS = 2
       H2 = FUNC_G(IS,MUt,NORD,CN_G)    ! STTAE IS = 2

       EX1 = (1-ETA)/(1-GAM)
       EX2 = (1-RHO)/(1-ETA)
       EX3 = 1/(1-RHO)
       EX4 = (1-RHO)/(1-GAM)
       H21 = H2/H1
       RHSB1 = ((1-BET)+BET*H1**EX4*(MUt+(1-MUt)*H21**EX1)**EX2)**EX3   !  RHS OF (B.1)

       RC(I) = RHSB1 - CBSVN(NORD,CN_G,2*MUt-1)          ! G(mu_t) = CBSVN(N,CN,2*PIt-1)
      ENDDO

      RETURN
      END
C========================================================================================
C THIS FUNCTION COMPUTES THE INNER EXPECTATION IN (B.1), USING QUDRATURE METHOD. THE
C EXPECTATION IS 2D BECAUSE mu_{t+1} DEPENDS ON eps_c AND eps_d.
C========================================================================================
      FUNCTION FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)
      IMPLICIT NONE
      INTEGER NQUD,NORD,IS,IC,ISTATE
      PARAMETER (nqud=10)
      REAL*8 FUNC_RM,MUt,CN_G(0:100),CN_RM(0:100)
      REAL*8 Zc,Zd,QWC,QWD,MUtp1,GMUtp1,CBSVN,Gc,Gd,PhiMUt,PI,
     &       KERNEL,PKERNEL,TEMP,RM,DR,HS(2),FUNC_G
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      ISTATE = 1
      HS(1) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 1
      ISTATE = 2
      HS(2) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 2

      PhiMUt = CBSVN(NORD,CN_RM,2*MUt-1)

      FUNC_RM = 0.0D0
      DO IC = 1, NQUD
       Zc = QX(IC)
       QWC = QW(IC)
        KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        TEMP = (1+CBSVN(NORD,CN_RM,2*MUtp1-1))/PhiMUt
        Gd = KaD(IS) + SigD*DSQRT(2.0D0)*Zc
        RM = TEMP*DEXP(Gd)       ! RM = TEMP*DEXP(Gd), Gd=KaD(IS)+SigD*DSQRT(2.0D0)*QXD
        FUNC_RM = FUNC_RM + KERNEL*RM*QWC
      ENDDO
      PI = 2*DASIN(1.0D0)
      FUNC_RM = FUNC_RM*DEXP(0.5*SigU**2)/DSQRT(PI)  ! DIVIDE SQRT(PI), COMING FROM 2D STD NORMAL DEN FUNCTION

      RETURN
      END
C========================================================================================
C THIS FUNCTION COMPUTES THE INNER EXPECTATION IN (B.1), USING QUDRATURE METHOD. THE
C EXPECTATION IS 2D BECAUSE mu_{t+1} DEPENDS ON eps_c AND eps_d.
C========================================================================================
      FUNCTION FUNC_G(IS,MUt,NORD,CN_G)
      IMPLICIT NONE
      INTEGER NQUD,NORD,IS,IC
      PARAMETER (nqud=10)
      REAL*8 FUNC_G,MUt,CN_G(0:100)
      REAL*8 QXC,QXD,QWC,QWD,MUtp1,GMUtp1,CBSVN,Gc,DR,PI
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      FUNC_G = 0.0D0
      DO IC = 1, NQUD
       QXC = QX(IC)
       QWC = QW(IC)
        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        GMUtp1 = CBSVN(NORD,CN_G,2*MUtp1-1)
        Gc = KaC(IS) + SigC*DSQRT(2.0D0)*QXC
        FUNC_G = FUNC_G + (GMUtp1*DEXP(Gc))**(1-GAM)*QWC    ! GUANSIAN 2D QUDRATURE
      ENDDO
      PI = 2*DASIN(1.0D0)
      FUNC_G = FUNC_G/DSQRT(PI)   ! DIVIDE SQRT(PI), COMING FROM 2D STD NORMAL DEN FUNCTION

      RETURN
      END
C=================================================================================
      FUNCTION BAYSIAN(IS,MUt,EPSc)
      IMPLICIT NONE
      INTEGER IS
      REAL*8 BAYSIAN,MUt,EPSc
      REAL*8 Gc,Gd,DET,C1,C2,D1,D2,EXPONT1,EXPONT2,FY21
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU

C--FACTOR SQRT(2) IS INTRODUCED TO CHANGE VARIABLE FROM NORMAL DENSITY FACTOR: EXP(X^2/),
C--TO A GUASIAN QUADRATURE FACTOR: EXP(-X^2). EXCEPT THE FACTOR SQRT(2), Zc AND Zd
C--ARE INDEPENDENT STD NORMAL. THE TRANSFORMATION IN Gc AND Gd YIELDS THE CORRECT
C--CORRELATION MATRIX. Zc AND Zd ARE Guassian quadrature points. INTEGRATION
C--(EXPECTATION) IS PERFORMED OVER THEM.

      Gc = KaC(IS) + SigC*EPSc     ! GROWTH RATES FROM STATE: IS (=1 OR 2)

C--EXPONENT OF THE EXPONENTIAL FUNCTION IN THE 2D NORMAL DENSITY FUNCTION

      EXPONT1 = -(Gc-KaC(1))**2/(2*SigC**2)
      EXPONT2 = -(Gc-KaC(2))**2/(2*SigC**2)

      fy21 = DEXP(EXPONT2-EXPONT1)         ! ratio = fy(y,2)/fy(y,1)

C--BASIAN UPDATING: (13)

      BAYSIAN = (Q*MUt + (1-P)*fy21*(1-MUt))/(MUt+fy21*(1-MUt))    ! BAYESIAN UPDATE OF MU_{t+1}(y)

      RETURN
      END
C=================================================================================
C RETURN THE BAYSIAN UPDATED DENSITY, GIVEN DRIFT MU(1), MU(2) AND DIFFUSION SIGMA.
C INPUTS: Y---OBSERVED DIV AT $t+1$; PIT = Pi_t---Pr(S_{t+1}=1|F_t); MU---DIMENSION
C         OF 2, MU(1), MU(2), DRIFT OF DIV OF TWO STATES = {1, 2}; SIG---VOLATILITY;
C         P---TRANSITION PROB OF STATE 1 TO STATE 1; Q---TRAN. PROB OF STATE 2 TO 2;
C OUTPUTS: PITP1 = Pi_{t+1}---Pr(S_{t+2}=1|F_{t+1}), BAYESIAN UPDATED PROB GIVEM Y.
C=================================================================================
      SUBROUTINE DEN_RATIO_GRID(NQUD,QX,FY21)
      IMPLICIT NONE
      INTEGER NQUD,IS,IC
      REAL*8 QX(NQUD),FY21(2,NQUD)
      REAL*8 Zc,Zd,Gc,Gd,DET,C1,C2,D1,D2,EXPONT1,EXPONT2
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU

C--FACTOR SQRT(2) IS INTRODUCED TO CHANGE VARIABLE FROM NORMAL DENSITY FACTOR: EXP(X^2/),
C--TO A GUASIAN QUADRATURE FACTOR: EXP(-X^2). EXCEPT THE FACTOR SQRT(2), Zc AND Zd
C--ARE INDEPENDENT STD NORMAL. THE TRANSFORMATION IN Gc AND Gd YIELDS THE CORRECT
C--CORRELATION MATRIX. Zc AND Zd ARE Guassian quadrature points. INTEGRATION
C--(EXPECTATION) IS PERFORMED OVER THEM.

      DO IS = 1, 2
       DO IC = 1, NQUD
        Zc = QX(IC)
        Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc     ! GROWTH RATES FROM STATE: IS (=1 OR 2)

C--EXPONENT OF THE EXPONENTIAL FUNCTION IN THE 2D NORMAL DENSITY FUNCTION
        EXPONT1 = -(Gc-KaC(1))**2/(2*SigC**2)
        EXPONT2 = -(Gc-KaC(2))**2/(2*SigC**2)

        FY21(IS,IC) = DEXP(EXPONT2-EXPONT1)         ! ratio = fy(y,2)/fy(y,1)
       ENDDO
      ENDDO

      RETURN
      END
C====================================================================
C GIVEN THE ORDER OF APPROXIMATION: NORD, AND THE COEFFICIENTS: CN(0:NORD),
C RETURN THE CHEBYSHEV POLYNORMIAL APPROXIMATION OF A FUNCTION.
C INPUTS: NORD---ORDER OF APPROXINATION;
C         CN(0:NORD)---APPROXIMATION COEF'S; INITIAL ORDER = 0; NORD <= 100;
C         THE REMAINING COEF'S FROM: (NORD+1):100, SET TO ZERO AND NOT USED.
C         Z---THE POINT AT WHICH THE VALUE OF THE FUNCTION IS APPROXIMATED;
C OUTPUTS: VALUE OF FUNCTION AT Z, WHICH IS [-1,1].
C====================================================================
      FUNCTION CBSVN(NORD,CN,Z)
      IMPLICIT NONE
      INTEGER NORD,I
      REAL*8 CBSVN,Z,C0,C1,C2,CN(0:100)

      IF (NORD.LT.2.OR.NORD.GT.100) STOP 'NEED 100 >= NORD >= 2'

      C0 = 1.0D0                       ! C0, C1: FIRST TWO CHEBYSHEV
      C1 = Z
      CBSVN = C0*CN(0) + C1*CN(1)      ! CBSVN = SUM_{I=0}^N T(I,Z)*CN(I)
      DO I = 2, NORD
       C2 = 2*Z*C1 - C0              ! RECURSIVELY DEFINED CHEBYSHEV T(N,Z)=2*Z*T(N-1,Z)-T(N-2,Z)
       CBSVN = CBSVN + C2*CN(I)      ! CHEBYSHEV APPROX WITH COEF COEF(0:N)
       C0 = C1                       ! UPDATE C0, C1 FOR NEXT ITERATION/LOOP
       C1 = C2
      ENDDO

      RETURN
      END
C========================================================================
C SOLVE THE FULL INFORMATION MODEL FOR THE VALUE FUNCTION; G(1) AND G(2).
C========================================================================
      SUBROUTINE FULL_REQS_G(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ
      REAL*8 SN(NEQ),RC(NEQ)
      REAL*8 EX1,EX2,EX3,H1,H2
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      EX1 = 1 - GAM
      EX2 = (1-RHO)/(1-GAM)
      EX3 = 1 - RHO

      H1 = SN(1)**EX1*DEXP(KaC(1)*EX1+0.5*EX1**2*SigC**2)
      H2 = SN(2)**EX1*DEXP(KaC(2)*EX1+0.5*EX1**2*SigC**2)

      RC(1) = (1-BET) + BET*(Q*H1+(1-Q)*H2)**EX2 - SN(1)**EX3
      RC(2) = (1-BET) + BET*((1-P)*H1+P*H2)**EX2 - SN(2)**EX3

      RETURN
      END
C========================================================================
C GIVEN THE VALUE FUNCTION: G(1) AND G(2), SOLVE THE FULL INFORMATION MODEL
C FOR THE PRICE DIV RATIO: varphi(1) AND varphi(2).
C========================================================================
      SUBROUTINE FULL_REQS_RM(G1,G2,VP1,VP2)
      IMPLICIT NONE
      REAL*8 G1,G2,VP1,VP2,COR
      REAL*8 EX1,EX2,EX3,VAR,CD1,CD2,C1,C2,DEN1,DEN2,X1,X2,Y1,Y2,DET
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      EX1 = 1 - GAM
      EX2 = (RHO-GAM)/(1-GAM)
      EX3 = RHO - GAM

      VAR = GAM**2*SigC**2 + SigD**2 - 2*GAM*SigC*SigD
      CD1 = DEXP(-GAM*KaC(1)+KaD(1)+0.5*VAR)
      CD2 = DEXP(-GAM*KaC(2)+KaD(2)+0.5*VAR)
      C1 = DEXP(KaC(1)*EX1+0.5*EX1**2*SigC**2)
      C2 = DEXP(KaC(2)*EX1+0.5*EX1**2*SigC**2)
      DEN1 = (Q*G1**EX1*C1+(1-Q)*G2**EX1*C2)**EX2
      DEN2 = ((1-P)*G1**EX1*C1+P*G2**EX1*C2)**EX2

      X1 = BET*Q*G1**EX3*CD1*DEXP(0.5*SigU**2)/DEN1
      Y1 = BET*(1-Q)*G2**EX3*CD2*DEXP(0.5*SigU**2)/DEN1
      X2 = BET*(1-P)*G1**EX3*CD1*DEXP(0.5*SigU**2)/DEN2
      Y2 = BET*P*G2**EX3*CD2*DEXP(0.5*SigU**2)/DEN2

      DET = (X1-1)*(Y2-1) - X2*Y1
      VP1 = -(Y2-Y1-1)/DET - 1
      VP2 = -(X1-X2-1)/DET - 1

      RETURN
      END
C==================================================================================
C RETURN A SERIES OF CONS GROWTH RATE, PRICE-DIV RATIO, AND EXCESS RETURN
C P=PR(S1=1|S0=1), Q=PR(S1=2|S0=2); THEN PB=1-P=PR(S1=2|S0=1); QB=1-Q=PR(S1=1|S0=2)
C=================================================================================
      SUBROUTINE FULL_MODEL_SERIES(CONS,PD,EXRET)       ! FULL MODEL PREDICTABILITY
      IMPLICIT NONE
      INTEGER NEQ,NORD,NR,NB,NC,I,J,PriS,NextS,N,IS,ISEED,IK
      PARAMETER (NEQ=21,NR=100000)
      COMMON /ORDER_STATE/N,IS
      REAL*8 RF,RE,CONS(NR),PD(NR),EXRET(NR),Prs,Pr,EPSILON,PI0,PI1,
     &       Yt,PHI0,PHI1,PIT,BAYSIAN,CBSVN,XVAR,EPSc,EPSd,RISKFREE
      REAL*8 DRNUNF,DRNNOF
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 CN_G(0:100),CN_RM(0:100)
      COMMON /G_RM_COEF/CN_G,CN_RM

      CALL FULL_MODEL_COEF(CN_G,CN_RM)       ! RETURN COEF TO A COMMON BLOCK

      NORD = NEQ - 1

      ISEED = 123457                    ! SET INITIAL SEED NUMBER FOR RANDOM GENERATOR
      CALL RNSET(ISEED)                 ! IMSL ROUTINE INITIALIZING SEED

      Prs = (1-P)/(2-P-Q)        ! UNCOND. PR(S=1), STATIONARY PROB OF BAD STATE
      PI0 = Prs

      Pr = DRNUNF()             ! RETURN A UNIFORM RANDOM NUMBER IN (0,1), SI=INITIAL STATE
      PriS = 2                  ! INITIAL DEFAULT STATE = STATE 2  (GOOD STATE)
      IF (Pr.LT.Prs) PriS = 1   ! DETERMINE INITIAL STATE, PR = PROB OF STATE 1 (BAD STATE)
      PI0 = PRS                 ! PRIOR PROB

      DO I = 1, NR
        Pr = DRNUNF()          ! DRAW ANOTHER UNIFORM [0,1] TO DETERMINE NEXT STATE
        NextS = 2              ! DEFAULT = STATE 2
        IF (PriS.EQ.2.AND.Pr.LT.1-P.OR.PriS.EQ.1.AND.Pr.LT.Q) NextS = 1
        EPSc = DRNNOF()               ! IMSL ROUTINE FOR STANDARD NORMAL. CORRESPONDING REALIZATION OF CONS
        EPSd = DRNNOF()               ! ANOTHER IID NORMAL FOR THE DIV PROCESS
        PIT = PI0                     ! PASS PI0 TO THE COMMON BLOCK FOR OTHER ROUTINES TO ACESS
        PI1 = BAYSIAN(NextS,PIt,EPSc)   ! BAYESIAN UPDATE
        RF = RISKFREE(NORD,PIT,CN_G)    ! GIVEN, PI_t, R^f_t COMPUTED
        PHI0 = CBSVN(NORD,CN_RM,2*PI0-1)                 ! GIVEN PI_t and Yt
        PD(I) = DLOG(PHI0)                   !     LOG(P/D)
        PHI1 = CBSVN(NORD,CN_RM,2*PI1-1)                 ! UPDATE PI_{t+1}(Yt;PI_t)
        CONS(I) = KaC(NextS) + SigC*EPSc
        Yt = KaD(NextS) + SigD*EPSc + SigU*EPSd
        RE = ((1+PHI1)/PHI0)*DEXP(Yt)      ! RE=(1+PHI(PI_{t+1})/PHI(PI_t)*EXP(Yt)
        EXRET(I) = DLOG(RE/RF)             ! LOG(RE/RF)
        PriS = NextS           ! UPDATE PRIOUS STATE, NEXT STATE BECOMES PREVIOUS STATE
        PI0 = PI1              ! PREVIOUS UPDATED PROB BECOMES THE INITIAL PROB FOR BEXT UPDATING
       ENDDO      ! END I LOOP

      RETURN
      END
C=====================================================================================
C GIVEN SERIES, COMPUTE CORRELATION OR AUTO CORRELATION
C=====================================================================================
      SUBROUTINE CORRELATION(CONS,PD,EXRET,COR)
      IMPLICIT NONE
      INTEGER NR,I
      PARAMETER (NR=100000)
      REAL*8 CONS(NR),PD(NR),EXRET(NR),COR(10),M(10),V(10),C(10)

      DO I = 1, 10
       M(I) = 0.0D0         ! FOR COMPUTING MEAN
       V(I) = 0.0D0         ! FOR COMPUTING VALRIANCE
       C(I) = 0.0D0         ! FOR COMPUTING CROSS PRODUCT
      ENDDO

      DO I = 1, NR-1
       M(1) = M(1) + CONS(I)         ! CONS
       M(2) = M(2) + PD(I)           ! PD
       M(3) = M(3) + EXRET(I+1)      ! Ex Ret & (t+1)
       M(4) = M(4) + DABS(EXRET(I))  ! ABS(Ex Ret & t)
       M(5) = M(5) + DABS(EXRET(I+1))  ! ABS(Ex Ret & (t+1))

       V(1) = V(1) + CONS(I)**2         ! CONS
       V(2) = V(2) + PD(I)**2           ! PD
       V(3) = V(3) + EXRET(I+1)**2      ! Ex Ret & (t+1)
       V(4) = V(4) + DABS(EXRET(I))**2  ! ABS(Ex Ret & t)
       V(5) = V(5) + DABS(EXRET(I+1))**2  ! ABS(Ex Ret & (t+1))

       C(1) = C(1) + CONS(I)*PD(I)       ! CONS*PD
       C(2) = C(2) + CONS(I)*EXRET(I+1)     ! CONS*Ex Ret & (t+1)
       C(3) = C(3) + CONS(I)*DABS(EXRET(I))     ! CONS*ABS(Ex Ret & t)
       C(4) = C(4) + DABS(EXRET(I))*DABS(EXRET(I+1))     ! ABS(Ex Ret & t)*ABS(Ex Ret & t+1)
       C(5) = C(5) + DABS(EXRET(I+1))*PD(I)        !  DABS(EXRET(I+1))*PD(I)
      ENDDO

      DO I = 1, 10
       M(I) = M(I)/(NR-1)         ! MEAN
       V(I) = V(I)/(NR-1)         ! MEAN OF SQUARE
       C(I) = C(I)/(NR-1)         ! MEAN OF CROSS PRODUCT
      ENDDO

      COR(1) = (C(1)-M(1)*M(2))/DSQRT((V(1)-M(1)**2)*(V(2)-M(2)**2))
      COR(2) = (C(2)-M(1)*M(3))/DSQRT((V(1)-M(1)**2)*(V(3)-M(3)**2))
      COR(3) = (C(3)-M(1)*M(4))/DSQRT((V(1)-M(1)**2)*(V(4)-M(4)**2))
      COR(4) = (C(4)-M(4)*M(5))/DSQRT((V(4)-M(4)**2)*(V(5)-M(5)**2))
      COR(5) = (C(5)-M(2)*M(5))/DSQRT((V(2)-M(2)**2)*(V(5)-M(5)**2))

      RETURN
      END
