C====================================================================
C WRITTEN BY NENGJIU JU AND JIANJUN MAIO
C
C THIS PROGRAM PRODUCES TABLE 3 IN JU AND MIAO (2010). IT CALLS IMSL
C FORTRAN ROUTINES. THESE ROUTINES CAN BE REPLACED BY ANOTHER LIBRARY
C ROUTINES OR BY SOURCE ROUTINES IF GOOD ONES ARE AVAILABLE.
C
C====================================================================
      USE NUMERICAL_LIBRARIES
      IMPLICIT NONE
      INTEGER NQUD,IW,NFIX,I,J,K,IP(6),NR,NB,ISEED,NEQ,NORD,ITER,NPAR
      PARAMETER (NEQ=21,NQUD=10,IW=4,NFIX=0,NR=10000,NB=500,NPAR=15)
      INTEGER STATES(NR)
      REAL*8 DEFAULTA,DEFAULTB,QXFIX(0:NFIX),SMUt(NR),STI(NR),
     &       STF(NR,NB),DIVGR(NR,NB),F_I_MNT(7),F_M_MNT(7),
     &       L_N_MNT(7),DCOMP(9),PR,EC,ED,ESigD,Gd,LAM,CBSVN
C      REAL*8 DRNUNF,DRNNOF        ! ON SOME COMPUTER SYSTEM, THE COMMENT NEEDS TO BE REMOVED
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21
      REAL*8 CN_G(0:100),CN_RM(0:100)
      COMMON /G_RM_COEF/CN_G,CN_RM
      REAL*8 FV_G(0:100),FV_RM(0:100)
      COMMON /FV_COEF/FV_G,FV_RM
      REAL*8 G1,G2,VP1,VP2
      COMMON /FULL_G_RM_COEF/G1,G2,VP1,VP2

      ISEED = 123457                    ! SET INITIAL SEED NUMBER FOR RANDOM GENERATOR
      CALL RNSET(ISEED)                 ! IMSL ROUTINE INITIALIZING SEED

C GENERATE THE RANDOM NUMEBRS TO BE USED REPEATEDLY FOR MULTIPLE INPUT SETS

      DO I = 1, NR
       STI(I) = DRNUNF()               ! RETURN A UNIFORM RANDOM NUMBER IN (0,1), SI=INITIAL STATE
       DO J = 1, NB
        STF(I,J) = DRNUNF()            ! DRAW ANOTHER UNIFORM [0,1] TO DETERMINE NEXT STATE
        DIVGR(I,J) = DRNNOF()          ! IMSL ROUTINE FOR STANDARD NORMAL.
       ENDDO      ! END J LOOP
      ENDDO       ! END I LOOP

C--COMUPTE THE GAUSSIAN QUADRATURE POINTS AND WEIGHTS

      CALL ERSET(3,0,-1)                   ! TURNING OFF WARNING MESSAGES

      DEFAULTA= 2.0D0   ! DEFAULTA AND DEFAULTB ARE IGNORED WHEN IW = 4 (THE CASSE HERE)
      DEFAULTB = 2.0D0
      CALL DGQRUL(NQUD,IW,DEFAULTA,DEFAULTA,NFIX,QXFIX,QX,QW)    ! QX CONTAINS POINTS, QW CONTAINS WEIGHTS

C--PARAMETERS OF DATA (CONSUMPTION) GENERATING PROCESS FROM cECCHETTI ET AL (2000, AER)

      Q = 0.516D0           ! BAD-TO-BAD STATE (STATE 1)
      P = 0.978D0           ! GOOD-TO-GOOD STATE (STATE 2)
      KaC(1) = -0.06785D0   ! BAD STATE CONS. GROWTH RATE
      KaC(2) = 0.02251D0    ! GOOD STATE CONS. GROWTH RATE
      SigC = 0.03127D0      ! STD OF CONS GROWTH RATE

      PR = (1-P)/(2-P-Q)    ! STATIONARY PROB OF BAD STATE = 1
      EC = PR*KaC(1) + (1-PR)*KaC(2)          ! EXPECTED CONS GROWTH RATE = 1.86%

C--PARAMETERS OF DATA DIVIDEND PROCESS

      LAM = 2.74D0          ! LEVERAGE EFFECT COEFFICIENT
      ESigD = 0.12D0        ! STD OF DIV CONDITION ON STATE, IMPLYING UNCOND STD = 0.13019 = 13%

      Gd = EC - LAM*EC           ! TO MAKE SURE E[DIV GROWTH RATE] = CONS GROWTH = EC = 1.86%.
      KaD(1) = LAM*KaC(1) + Gd   ! BAD STATE DIV. GROWTH RATE
      KaD(2) = LAM*KaC(2) + Gd   ! GOOD STATE DIV. GROWTH RATE
      SigD = LAM*SigC            ! STD OF THE LEVERED PART OF THE DIV PROCESS IN (14)
      SigU = DSQRT(ESigD**2-LAM**2*SigC**2)  ! THIS IS SigD IN (14) NAMED DIFFERENTLY = 0.084018

C--OPEN FILES FOR INPUT AND OUTPUT

      OPEN (UNIT=2, FILE='Table3And4Input.dat')
      OPEN (UNIT=3, FILE='Table3Output.dat')
      OPEN (UNIT=4, FILE='Table4Output.dat')

C--PREFERENCE MODEL PARAMETERS

      DO I = 1, NPAR
        WRITE(6,*) I                 ! MONITORING PROGRAM PROGRESS 
        READ(2,*) BET,RHO,GAM,ETA    ! ONE PARAM SET, TOTAL OF NPAR SETS

      IF (ETA.LT.GAM) STOP 'ETA HAS TO >= GAM'

C--GIVEN DATA DYNAMICS, SIMULATE STATE AND SAVE FOR REPEATED USE

      CALL SIMUSTATE(NR,NB,STI,STF,DIVGR,P,Q,SMUt,STATES)

C--RETURN THE DENSITY RATIOS FOR A GIVEN PAIR (CONS RATE, DIV RATE) USED IN BAYSIAN
C--UPDATING TO A COMMON BLOCK, USED REPEATEDLY

      CALL DEN_RATIO_GRID(NQUD,QX,FY21)

C--FULL INFORMATION BUT NO AMBIGUITY MODEL

      CALL FULL_INFO_MOMENTS(NR,STATES,F_I_MNT)

C--FULL MODEL WITH LEARNING AND AMBIGUITY

      CALL FULL_MODEL_MOMENTS(NR,SMUt,F_M_MNT)

      WRITE(3,42) F_M_MNT*100
42    FORMAT(1X,7F7.2)

C--LEARNING MODEL WITHOUT AMBIGUITY

      ETA = GAM              ! NOW AMBIGUITY NEUTRAL (NO AMBIGUITY)
      CALL FULL_MODEL_MOMENTS(NR,SMUt,L_N_MNT)

      CALL DECOMP(F_I_MNT,F_M_MNT,L_N_MNT,DCOMP)

      WRITE(4,44) DCOMP*100
44    FORMAT(1X,9F7.2)

      ENDDO  ! END I LOOP OF ONE PARAMETER SET
      CLOSE(2)
      CLOSE(3)
      CLOSE(4)

      STOP
      END
C=====================================================================================
C GIVEN THE FIRST MOMENTS OF Rf, Re, AND PREMIUM, THE ROUTINE DECOMPOSES FULL MOMENTS
C INTO COMPONENTS DUE (1) FULL INFO; (2) LEARNING; (3) AMBIGUITY.
C=====================================================================================
      SUBROUTINE DECOMP(F_I_MNT,F_M_MNT,L_N_MNT,DCOMP)
      IMPLICIT NONE
      REAL*8 F_I_MNT(7),F_M_MNT(7),L_N_MNT(7),DCOMP(9)

      DCOMP(1) = F_I_MNT(1)                   ! FIRST MOMENTS FROM FULL INFO MODEL
      DCOMP(2) = L_N_MNT(1) - F_I_MNT(1)      ! Rf^L - Rf^F, LEARNING BUT NO AMBIGUITY - FULL MODEL VALUE
      DCOMP(3) = F_M_MNT(1) - L_N_MNT(1)      ! Rf - Rf^L, FULL MODEL - LEARNING BUT NO AMBIGUITY

      DCOMP(4) = F_I_MNT(3)                   ! EQUITY RETURN DECOMPOSITION
      DCOMP(5) = L_N_MNT(3) - F_I_MNT(3)
      DCOMP(6) = F_M_MNT(3) - L_N_MNT(3)

      DCOMP(7) = F_I_MNT(6)                   ! EQUITY PREMIUM DECOMPOSITION
      DCOMP(8) = L_N_MNT(6) - F_I_MNT(6)
      DCOMP(9) = F_M_MNT(6) - L_N_MNT(6)

      RETURN
      END
C=====================================================================================
C RETURN UNCONDITIONAL MOMENTS OF THE LEARNING AND AMBIGUITY MODEL, WHICH INCLUDES
C LEARNING BUT WITHOUT AMBIGUITY MODEL AS A SPECIAL CASE BY SETTING: ETA = GAMMA
C=====================================================================================
      SUBROUTINE FULL_MODEL_MOMENTS(NR,SMUt,F_M_MNT)
      IMPLICIT NONE
      INTEGER NR,NEQ,NORD,SNR,IP(6),I
      PARAMETER (NEQ=21,SNR=10000)
      REAL*8 SMUt(NR),MNPI(SNR,8),F_M_MNT(7)
      REAL*8 XG(NEQ),XS(NEQ),FS(NEQ),RP(5),X(NEQ),FV(NEQ)
      DATA XS/NEQ*1.0D0/,FS/NEQ*1.0D0/         ! using DEFAULT VALUES AND INITIALIZE CN TO 0
      REAL*8 CN_G(0:100),CN_RM(0:100)
      COMMON /G_RM_COEF/CN_G,CN_RM
      REAL*8 FV_G(0:100),FV_RM(0:100)
      COMMON /FV_COEF/FV_G,FV_RM
      REAL*8 G1,G2,VP1,VP2
      COMMON /FULL_G_RM_COEF/G1,G2,VP1,VP2
      EXTERNAL REQS_G,REQS_RM

      XG(1) = G2             ! SET CN(0) TO FULL-INFO MODEL HIGH STATE VALUE
      DO I = 2, NEQ          ! SET OTHERS INITIALLY TO ZERO
       XG(I) = 0.0D0
      ENDDO

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(REQS_G,NEQ,XG,XS,FS,IP,RP,X,FV)   ! IMSL ROUTINE SOLVING nonlinear equations

C--FINAL CN(0:NORD) COEFFICIENTS

      NORD = NEQ - 1
      DO I = 0, NORD
       CN_G(I) = X(I+1)
       FV_G(I) = FV(I+1)
      ENDDO

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR \varphi(mu_t).

      XG(1) = VP2            ! INITIAL GUESS OF PRICE/DIV RATIO, SET TO FULL-INFO MODEL HIGH STATE VALUE
      DO I = 2, NEQ                ! SET OTHERS INITIALLY TO ZERO
       XG(I) = 0.0D0
      ENDDO

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(REQS_RM,NEQ,XG,XS,FS,IP,RP,X,FV)  ! IMSL ROUTINE SOLVING nonlinear equations

C--FINAL CN_RM(0:NORD) COEFFICIENTS

      NORD = NEQ - 1
      DO I = 0, NORD
       CN_RM(I) = X(I+1)
       FV_RM(I) = FV(I+1)
      ENDDO

C--SIMULATE MUt AND RETURN CONDITIONAL MOMENTS

      CALL SIMURET(NORD,NR,CN_G,CN_RM,SMUt,MNPI)

C--GIVEN CONDITIONAL MOMENTS: MNPI, RETURN UNCONDITIONAL MOMENTS: MNT

      CALL RetSTATICS(NR,MNPI,F_M_MNT)

      RETURN
      END
C=====================================================================================
C RETURN UNCONDITIONAL MOMENTS OF THE FULL INFORMATION MODEL
C=====================================================================================
      SUBROUTINE FULL_INFO_MOMENTS(NR,STATES,F_I_MNT)
      IMPLICIT NONE
      INTEGER NR,NEQ,SNR,IP(6)
      PARAMETER (NEQ=2,SNR=10000)
      INTEGER STATES(NR)
      REAL*8 MNPI(SNR,8),F_I_MNT(7),GTC,G0
      REAL*8 XG(NEQ),XS(NEQ),FS(NEQ),RP(5),X(NEQ),FV(NEQ)
      DATA XS/NEQ*1.0D0/,FS/NEQ*1.0D0/          ! using DEFAULT VALUES AND INITIALIZE CN TO 0
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 G1,G2,VP1,VP2
      COMMON /FULL_G_RM_COEF/G1,G2,VP1,VP2
      EXTERNAL FULL_REQS_G

C--CHECK TRANSVERSALITY CONDITION. ASSUME G(Z_t) IS A CONSTANT AND SOLVE THE
C--OBJECTIVE FUNCTION (RECURSIVE) EQUATION FOR AN INITIAL GUESS. ASSUME STATE
C--AT t+1 IS 2 (HIGH GROWTH STATE).

      GTC = BET*DEXP((KaC(2)+0.5*(1-GAM)*SigC**2)*(1-RHO))
      IF (GTC.GE.1.0D0) THEN
       WRITE(6,*) 'GTC ',GTC
       WRITE(6,*) 'MAY VIOLATE TRANS. CONDITION. CHECK TO MAKE SURE'
       STOP
      ENDIF

      G0 = 1.0D0
      IF (GTC.LT.1.0D0) G0 = ((1-BET)/(1-GTC))**(1/(1-RHO))   ! IF GTC < 1, TRANSVERSALITY CONDITION OK

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR G(Z_t): G(1) AND G(2).

      XG(1) = G0                   ! SET INITIAL GUESS TO G0
      XG(2) = G0

      IP(1) = 0                ! USE DEFAULT VALUES FOR IP(1:6) AND RP(1:5)
      CALL DNEQBF(FULL_REQS_G,NEQ,XG,XS,FS,IP,RP,X,FV)   ! IMSL ROUTINE SOLVING nonlinear equations
      G1 = X(1)
      G2 = X(2)

      IF (MIN(G1,G2).LT.0.0D0) THEN
       WRITE(6,*) 'G1, G2 ', G1,G2
       WRITE(6,*) 'G1, G2 HAS TO BE POTIVE'
       STOP
      ENDIF

C--IMSL ROUTINE SOLVING nonlinear equations WITH INITIAL GUESS FOR \varphi(Z_t): \varphi(1) AND
C--\varphi(2). ASSUME G(Z_t) IS A CONSTANT AND SOLVE THE EULAER EQUATION FOR AN INITIAL GUESS.
C--ASSUME STATE AT t+1 IS 2 (HIGH GROWTH STATE).

      CALL FULL_REQS_RM(G1,G2,VP1,VP2)            ! solve P/D ratio (VP1, VP2)

      IF (MIN(VP1,VP2).LT.0.0D0) THEN
       WRITE(6,*) 'VP1, VP2 ', VP1,VP2
       WRITE(6,*) 'P/D RATIO HAS TO BE POTIVE'
       STOP
      ENDIF

      CALL FULL_SIMURET(NR,BET,GAM,RHO,STATES,MNPI)

C--GIVEN CONDITIONAL MOMENTS: MNPI, RETURN UNCONDITIONAL MOMENTS: EMNPI

      CALL RetSTATICS(NR,MNPI,F_I_MNT)

      RETURN
      END
C=====================================================================================
C GIVEN MATRIX FOR Rf AND Re, COMPUTE SUMMARY STATISTICS BY TREATING N x M SIMULATIONS
C AS N SERIES.
C=====================================================================================
      SUBROUTINE RetSTATICS(NR,MNPI,MNT)
      IMPLICIT NONE
      INTEGER NR,I,J
      REAL*8 MNPI(NR,8),MNT(7),MN(8)

      DO J = 1, 8
       MN(J) = 0.0D0
      ENDDO

      DO J = 1, 8
       DO I = 1, NR              ! uncond. first and second moments
        MN(J) = MN(J) + MNPI(I,J)
       ENDDO
      ENDDO

      DO J = 1, 8
       MN(J) = MN(J)/NR
      ENDDO

      MNT(1) = MN(1) - 1                   ! UNCOND. MEAN Rf
      MNT(3) = MN(3) - 1                   ! UNCOND. MEAN Re
      MNT(5) = MN(5)                       ! UNCOND. MEAN OF (Re-Rf)
      MNT(2) = DSQRT(MN(2)-MN(1)**2)       ! UNCOND. VAR/STD OF Rf
      MNT(4) = DSQRT(MN(4)-MN(3)**2)       ! UNCOND. VAR/STD OF Re
      MNT(6) = DSQRT(MN(6)-MN(5)**2)       ! UNCOND. VAR/STD OF (Re-Rf)
      MNT(7) = DSQRT(MN(8)-MN(7)**2)/(100*MN(7))     ! STD(M)/E[M]

      RETURN
      END
C=================================================================================
C RETURN N SERIES OF STATES, EACH WITH A SIZE OF M. MARKOV STATE MVS(N,M).
C P=PR(S1=1|S0=1), Q=PR(S1=2|S0=2); THEN PB=1-P=PR(S1=2|S0=1); QB=1-Q=PR(S1=1|S0=2)
C=================================================================================
      SUBROUTINE SIMURET(NORD,NR,CN_G,CN_RM,MUt,MNPI)
      IMPLICIT NONE
      INTEGER NORD,NR,I,J
      REAL*8 CN_G(0:100),CN_RM(0:100),MUt(NR),MNPI(NR,8)
      REAL*8 PIT,RF,RE1,RE2,EM1,EM2

      DO I = 1, NR
       PIT = MUt(I)
       CALL COND_MOMENTS(NORD,CN_G,CN_RM,PIt,RF,RE1,RE2,EM1,EM2)

       MNPI(I,1) = RF                            ! COND. MEAN OF Rf ON MUt
       MNPI(I,2) = RF**2                         ! COND. SECOND MOMENT
       MNPI(I,3) = RE1                           ! COND. MEAN OF RE ON MUt
       MNPI(I,4) = RE2                           ! COND. SECOND MOMENT
       MNPI(I,5) = RE1 - RF                      ! COND. MEAN OF PRE (RE-RF) ON MUt
       MNPI(I,6) = RE2 + RF**2 - 2*RF*RE1        ! COND. SECOND OF (RE-RF)
       MNPI(I,7) = EM1                           ! COND. MEAN OF M_{t+1}
       MNPI(I,8) = EM2                           ! COND. SECOND MOMENT OF M_{t+1}
      ENDDO                ! END I LOOP

      RETURN
      END
C=================================================================================
C SIMULATE THE STATE AND SAVE AND CAN BE USED REPEATEDLY
C=================================================================================
      SUBROUTINE SIMUSTATE(NR,NB,STI,STF,DIVGR,P,Q,MUt,STATES)
      IMPLICIT NONE
      INTEGER NR,NB,I,J,PriS,NextS,STATES(NR)
      REAL*8 STI(NR),STF(NR,NB),DIVGR(NR,NB),P,Q,MUt(NR),Prs,Pr,
     &       PI0,PI1,PIT,BAYSIAN,EPSc

      Prs = (1-P)/(2-P-Q)        ! UNCOND. PR(S=1), STATIONARY PROB OF BAD STATE
      PI0 = Prs

      DO I = 1, NR
       Pr = STI(I)               ! RETURN A UNIFORM RANDOM NUMBER IN (0,1), SI=INITIAL STATE
       PriS = 2                  ! INITIAL DEFAULT STATE = STATE 2  (GOOD STATE)
       IF (Pr.LT.Prs) PriS = 1   ! DETERMINE INITIAL STATE, PR = PROB OF STATE 1 (BAD STATE)
       PI0 = PRS                 ! PRIOR PROB

       DO J = 1, NB            ! J-LOOP SIMULATES MARKOV CHAIN STATE AND BAYESIAN UPDATING
        Pr = STF(I,J)          ! DRAW ANOTHER UNIFORM [0,1] TO DETERMINE NEXT STATE
        NextS = 2              ! DEFAULT = STATE 2
        IF (PriS.EQ.2.AND.Pr.LT.1-P.OR.PriS.EQ.1.AND.Pr.LT.Q) NextS = 1
        EPSc = DIVGR(I,J)               ! IMSL ROUTINE FOR STANDARD NORMAL. CORRESPONDING REALIZATION OF CONS
        PIT = PI0                       ! PASS PI0 TO THE COMMON BLOCK FOR OTHER ROUTINES TO ACESS
        PI1 = BAYSIAN(NextS,PIt,EPSc)   ! BAYESIAN UPDATE
        PriS = NextS           ! UPDATE PREVIOUS STATE---CURRENT STATE BECOMES PREVIOUS STATE
        PI0 = PI1              ! PREVIOUS UPDATED PROB BECOMES THE INITIAL PROB FOR BEXT UPDATING
       ENDDO          ! END J LOOP
       MUt(I) = PIT
       STATES(I) = PriS
      ENDDO                ! END I LOOP

      RETURN
      END
C================================================================================================
C RETURN THE TWO FACTORS IN COMPUTING RISKFREE RATE AT EACH TIME POINT, GIVEN THE STATE PROB DIST.
C================================================================================================
      SUBROUTINE COND_MOMENTS(NORD,CN_G,CN_RM,MUt,RF,RE1,RE2,EM1,EM2)
      IMPLICIT NONE
      INTEGER NQUD,NORD
      REAL*8 CN_G(0:100),CN_RM(0:100),MUt,RF,RE1,RE2,EM1,EM2
      REAL*8 EM(2,2),RE(2,2)

C--COMPUTE COND. Rf AND MOMENTS OF PRICING KERNEL

      CALL M_MONTS(NORD,CN_G,CN_RM,MUt,EM,RE)

      RE1 = MUt*RE(1,1) + (1-MUt)*RE(2,1)  ! MEAN OF RM COND. ON MUt
      RE2 = MUt*RE(1,2) + (1-MUt)*RE(2,2)  ! MEAN OF RM^2

      EM1 = MUt*EM(1,1) + (1-MUt)*EM(2,1)  ! MEAN OF M_{t+1} COND. ON MUt
      EM2 = MUt*EM(1,2) + (1-MUt)*EM(2,2)  ! MEAN OF [M_{t+1}]^2
      RF = 1/EM1                           ! RISKFREE RATE

      RETURN
      END
C================================================================================================
C RETURN THE TWO FACTORS IN COMPUTING RISKFREE RATE AT EACH TIME POINT, GIVEN THE STATE PROB DIST.
C================================================================================================
      SUBROUTINE M_MONTS(NORD,CN_G,CN_RM,MUt,EM,RE)
      IMPLICIT NONE
      INTEGER NQUD,NORD,IC,ID,IS,ISTATE
      PARAMETER (nqud=10)
      REAL*8 CN_G(0:100),CN_RM(0:100),MUt,EM(2,2),RE(2,2)
      REAL*8 Zc,Zd,QWC,QWD,PKERNEL,KERNEL,PROD,HS(2),
     &       FUNC_G,DR,MUtp1,PhiMUtp1,PhiMUt,CBSVN,Gd,RM,PI
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      ISTATE = 1
      HS(1) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 1
      ISTATE = 2
      HS(2) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 2

      PhiMUt = CBSVN(NORD,CN_RM,2*MUt-1)            ! varphi(mu_t)

      DO IC = 1, 2           ! INITIALIZE TO ZERO
       DO ID = 1, 2
        EM(IC,ID) = 0.0D0
        RE(IC,ID) = 0.0D0
       ENDDO
      ENDDO

      DO IC = 1, NQUD
       Zc = QX(IC)
       QWC = QW(IC)
        IS = 1
        KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
        EM(1,1) = EM(1,1) + KERNEL*QWC       ! GUANSIAN 1D QUDRATURE, COMPUTE E[M]
        EM(1,2) = EM(1,2) + KERNEL**2*QWC    ! GUANSIAN 2D QUDRATURE, COMPUTE E[M^2]

        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        PhiMUtp1 = CBSVN(NORD,CN_RM,2*MUtp1-1)        ! varphi(mu_{t+1})
        Gd = KaD(IS) + SigD*DSQRT(2.0D0)*Zc
        RM = DEXP(Gd+0.5*SigU**2)*(1+PhiMUtp1)/PhiMUt       ! RM = GROWTH EQUITY RETURN

        RE(1,1) = RE(1,1) + RM*QWC
        RE(1,2) = RE(1,2) + RM**2*DEXP(SigU**2)*QWC   ! EXTRA EXP(SigU^2) FROM E[EXPONENTIAL NORMAL]

        IS = 2
        KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
        EM(2,1) = EM(2,1) + KERNEL*QWC       ! GUANSIAN 2D QUDRATURE
        EM(2,2) = EM(2,2) + KERNEL**2*QWC    ! GUANSIAN 2D QUDRATURE

        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        PhiMUtp1 = CBSVN(NORD,CN_RM,2*MUtp1-1)        ! varphi(mu_{t+1})
        Gd = KaD(IS) + SigD*DSQRT(2.0D0)*Zc
        RM = DEXP(Gd+0.5*SigU**2)*(1+PhiMUtp1)/PhiMUt       ! RM = GROWTH EQUITY RETURN

        RE(2,1) = RE(2,1) + RM*QWC
        RE(2,2) = RE(2,2) + RM**2*DEXP(SigU**2)*QWC   ! EXTRA EXP(SigU^2) FROM E[EXPONENTIAL NORMAL]

      ENDDO

      PI = 2*DASIN(1.0D0)
      DO IC = 1, 2           ! DIVIDE SQRT(PI), COMING FROM 1D STD NORMAL DEN FUNCTION
       DO ID = 1, 2
        EM(IC,ID) = EM(IC,ID)/DSQRT(PI)
        RE(IC,ID) = RE(IC,ID)/DSQRT(PI)
       ENDDO
      ENDDO

      RETURN
      END
C======================================================================
C RETURN THE PRICING KERNEL, GIVEN MUt,Zc,Zd, AND STATE IS = (1, OR 2),
C GROWTH RATES: Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc,
C Gd = KaD(IS) + SigD*DSQRT(2.0D0)*(COR*Zc+DSQRT(1-COR**2)*Zd).
C======================================================================
      FUNCTION PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
      IMPLICIT NONE
      INTEGER IS,IC,NQUD,NORD,ISTATE
      PARAMETER (nqud=10)
      REAL*8 PKERNEL,CN_G(0:100),MUt,HS(2),Zc
      REAL*8 Gc,FAC1,FAC2,FAC3,EX1,EX2,MUtp1,GMUtp1,CBSVN,
     &       Vtp1,RtVtp1,DR
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      Zc = QX(IC)

C--FIRST FACTOR: BET*Gc^{-RHO}

      Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc        ! LOG(C_{t+1}/C_t), CONS GROWTH RATE
      FAC1 = BET*DEXP(-RHO*Gc)

C--SECOND FACTOR: (Vtp1/RtVtp1)^(RHO-GAM).
C--HS(1:2) ARE THE TWO CONDITIONAL EXPECTATION IN Rt(V_{t+1}) IN (8). THE SAME
C--HS(1:2) (CONDITIONAL EXPECTATIONS) ARE USED IN (B.1) FOR THE VALUE FUNC.

      DR = FY21(IS,IC)                     !   DENSITY RATIO
      MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
      GMUtp1 = CBSVN(NORD,CN_G,2*MUtp1-1)       !   G(mu_{t+1})
      Vtp1 = GMUtp1*DEXP(Gc)                  !   V_{t+1}/C_t

      EX1 = (1-ETA)/(1-GAM)
      EX2 = 1/(1-ETA)
      RtVtp1 = (MUt*HS(1)**EX1+(1-MUt)*HS(2)**EX1)**EX2       ! Rt(v_{t+1})

      EX1 = RHO - GAM
      FAC2 = (Vtp1/RtVtp1)**EX1

C--THIRD FACTOR: (E_{t,J}[V_{t+1}^{1-GAM}])^{-(ETA-GAM)/(1-GAM)}/RtVtp1^{-(ETA-GAM)}

      EX1 = -(ETA-GAM)/(1-GAM)
      EX2 = -(ETA-GAM)
      FAC3 = HS(IS)**EX1/RtVtp1**EX2

C--CONDITIONAL PRICING KERNEL

      PKERNEL = FAC1*FAC2*FAC3

      RETURN
      END
C======================================================================
C RETURN NEQ RESIDUAL VALUES FOR THE SYSTEM OF NEQ EQUATIONS.
C
C INPUTS: NEQ---NUMBER OF RESIDUAL EQUATIONS = NUMBER OF COEF'S;
C         SN(NEQ)---COEF'S;
C         OTHER INPUT VALUES USED---PASSED THROUGH THE COMMON BLOCKS;
C
C OUTPUTS: RESIDUL(NEQ)---NEQ RESIDUALS, EACH FOR ONE PIt VALUE = NODAL VALUE OF CHEBYSHEV(NEQ)
C
C FUNCTIONS USED: CBSVN(N,CN,Z)---RETURN THE VALUE OF THE LAST TERM IN RESIDUAL R(PIt;CN);
C
C SUBROUTINE CALLED: DQDAGS---IMSL LIBRARY ROUTINE FOR INTEGRATION OF ONE-VARIABLE FUNCTION;
C
C======================================================================
      SUBROUTINE REQS_RM(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ,NORD,IS,I
      REAL*8 SN(NEQ),RC(NEQ)
      REAL*8 H1,H2,FUNC_RM,PI,MUt,RHSB2
      REAL*8 CN_G(0:100),CN_RM(0:100)
      COMMON /G_RM_COEF/CN_G,CN_RM

      NORD = NEQ - 1
      DO I = 0, NORD             ! PASS CN TO COMMON BLOCK, USED BY OTHER SUBROUTINES
       CN_RM(I) = SN(I+1)
      ENDDO

      PI = 2*DASIN(1.0D0)
      DO I = 1, NEQ               ! NEQ EQUATIONS TO DETERMINE C0,C1,....C(NEQ-1)
       MUt = 0.5 + 0.5*DCOS((NEQ-I+0.5)*PI/NEQ)    ! NEQ NODES OVER [0,1] OF CHEBYSHEV(NEQ)

C--RETURN THE INNER EXPECTATION IN (B.1) IS = 1, 2

       IS = 1
       H1 = FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)    ! STTAE IS = 1
       IS = 2
       H2 = FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)    ! STTAE IS = 2
       RHSB2 = MUt*H1 + (1-MUt)*H2       !  RHS OF (B.2)
       RC(I) = RHSB2 - 1           ! THE EULER EQUATION (B.3)
      ENDDO

      RETURN
      END
C========================================================================================
C RETURN NEQ RESIDUAL VALUES FOR THE SYSTEM OF NEQ EQUATIONS.
C
C INPUTS: NEQ---NUMBER OF RESIDUAL EQUATIONS = NUMBER OF COEF'S;
C         SN(NEQ)---COEF'S;
C         OTHER INPUT VALUES USED---PASSED THROUGH THE COMMON BLOCKS;
C
C OUTPUTS: RESIDUL(NEQ)---NEQ RESIDUALS, EACH FOR ONE PIt VALUE = NODAL VALUE OF CHEBYSHEV(NEQ)
C
C FUNCTIONS USED: CBSVN(N,CN,Z)---RETURN THE VALUE OF THE LAST TERM IN RESIDUAL R(PIt;CN);
C
C SUBROUTINE CALLED: DQDAGS---IMSL LIBRARY ROUTINE FOR INTEGRATION OF ONE-VARIABLE FUNCTION;
C
C======================================================================
      SUBROUTINE REQS_G(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ,NQUD,NORD,IS,I
      PARAMETER (nqud=10)
      REAL*8 SN(NEQ),RC(NEQ),CBSVN
      REAL*8 H1,H2,H21,FUNC_G,PI,MUt,CN_G(0:100),EX1,EX2,EX3,EX4,RHSB1
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      NORD = NEQ - 1
      DO I = 0, NORD             ! PASS CN TO COMMON BLOCK, USED BY OTHER SUBROUTINES
       CN_G(I) = SN(I+1)
      ENDDO

      PI = 2*DASIN(1.0D0)
      DO I = 1, NEQ               ! NEQ EQUATIONS TO DETERMINE C0,C1,....C(NEQ-1)
       MUt = 0.5 + 0.5*DCOS((NEQ-I+0.5)*PI/NEQ)    ! NEQ NODES OVER [0,1] OF CHEBYSHEV(NEQ)

C--RETURN THE INNER EXPECTATION IN (B.1) IS = 1, 2

       IS = 1
       H1 = FUNC_G(IS,MUt,NORD,CN_G)    ! STTAE IS = 1
       IS = 2
       H2 = FUNC_G(IS,MUt,NORD,CN_G)    ! STTAE IS = 2

       EX1 = (1-ETA)/(1-GAM)
       EX2 = (1-RHO)/(1-ETA)
       EX3 = 1/(1-RHO)
       EX4 = (1-RHO)/(1-GAM)
       H21 = H2/H1
       RHSB1 = ((1-BET)+BET*H1**EX4*(MUt+(1-MUt)*H21**EX1)**EX2)**EX3   !  RHS OF (B.1)

       RC(I) = RHSB1 - CBSVN(NORD,CN_G,2*MUt-1)          ! G(mu_t) = CBSVN(N,CN,2*PIt-1)
      ENDDO

      RETURN
      END
C========================================================================================
C THIS FUNCTION COMPUTES THE INNER EXPECTATION IN (B.1), USING QUDRATURE METHOD. THE
C EXPECTATION IS 2D BECAUSE mu_{t+1} DEPENDS ON eps_c AND eps_d.
C========================================================================================
      FUNCTION FUNC_RM(IS,MUt,NORD,CN_G,CN_RM)
      IMPLICIT NONE
      INTEGER NQUD,NORD,IS,IC,ISTATE
      PARAMETER (nqud=10)
      REAL*8 FUNC_RM,MUt,CN_G(0:100),CN_RM(0:100)
      REAL*8 Zc,Zd,QWC,QWD,MUtp1,GMUtp1,CBSVN,Gc,Gd,PhiMUt,PI,
     &       KERNEL,PKERNEL,TEMP,RM,DR,HS(2),FUNC_G
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      ISTATE = 1
      HS(1) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 1
      ISTATE = 2
      HS(2) = FUNC_G(ISTATE,MUt,NORD,CN_G)    ! STTAE IS = 2

      PhiMUt = CBSVN(NORD,CN_RM,2*MUt-1)

      FUNC_RM = 0.0D0
      DO IC = 1, NQUD
       Zc = QX(IC)
       QWC = QW(IC)
        KERNEL = PKERNEL(IS,NORD,CN_G,MUt,HS,IC)
        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        TEMP = (1+CBSVN(NORD,CN_RM,2*MUtp1-1))/PhiMUt
        Gd = KaD(IS) + SigD*DSQRT(2.0D0)*Zc
        RM = TEMP*DEXP(Gd)       ! RM = TEMP*DEXP(Gd), Gd=KaD(IS)+SigD*DSQRT(2.0D0)*QXD
        FUNC_RM = FUNC_RM + KERNEL*RM*QWC
      ENDDO
      PI = 2*DASIN(1.0D0)
      FUNC_RM = FUNC_RM*DEXP(0.5*SigU**2)/DSQRT(PI)  ! DIVIDE SQRT(PI), COMING FROM 1D STD NORMAL DEN FUNCTION

      RETURN
      END
C========================================================================================
C THIS FUNCTION COMPUTES THE INNER EXPECTATION IN (B.1), USING QUDRATURE METHOD. THE
C EXPECTATION IS 2D BECAUSE mu_{t+1} DEPENDS ON eps_c AND eps_d.
C========================================================================================
      FUNCTION FUNC_G(IS,MUt,NORD,CN_G)
      IMPLICIT NONE
      INTEGER NQUD,NORD,IS,IC
      PARAMETER (nqud=10)
      REAL*8 FUNC_G,MUt,CN_G(0:100)
      REAL*8 QXC,QXD,QWC,QWD,MUtp1,GMUtp1,CBSVN,Gc,DR,PI
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA
      REAL*8 QX(NQUD),QW(NQUD)
      COMMON /QuadPtrWght/QX,QW
      REAL*8 FY21(2,NQUD)
      COMMON /DEN_RATIO/FY21

      FUNC_G = 0.0D0
      DO IC = 1, NQUD
       QXC = QX(IC)
       QWC = QW(IC)
        DR = FY21(IS,IC)                     !   DENSITY RATIO
        MUtp1 = (Q*MUt + (1-P)*DR*(1-MUt))/(MUt+DR*(1-MUt))
        GMUtp1 = CBSVN(NORD,CN_G,2*MUtp1-1)
        Gc = KaC(IS) + SigC*DSQRT(2.0D0)*QXC
        FUNC_G = FUNC_G + (GMUtp1*DEXP(Gc))**(1-GAM)*QWC    ! GUANSIAN 2D QUDRATURE
      ENDDO
      PI = 2*DASIN(1.0D0)
      FUNC_G = FUNC_G/DSQRT(PI)   ! DIVIDE SQRT(PI), COMING FROM 2D STD NORMAL DEN FUNCTION

      RETURN
      END
C=================================================================================
      FUNCTION BAYSIAN(IS,MUt,EPSc)
      IMPLICIT NONE
      INTEGER IS
      REAL*8 BAYSIAN,MUt,EPSc
      REAL*8 Gc,Gd,DET,C1,C2,D1,D2,EXPONT1,EXPONT2,FY21
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU

C--FACTOR SQRT(2) IS INTRODUCED TO CHANGE VARIABLE FROM NORMAL DENSITY FACTOR: EXP(X^2/),
C--TO A GUASIAN QUADRATURE FACTOR: EXP(-X^2). EXCEPT THE FACTOR SQRT(2), Zc AND Zd
C--ARE INDEPENDENT STD NORMAL. THE TRANSFORMATION IN Gc AND Gd YIELDS THE CORRECT
C--CORRELATION MATRIX. Zc AND Zd ARE Guassian quadrature points. INTEGRATION
C--(EXPECTATION) IS PERFORMED OVER THEM.

      Gc = KaC(IS) + SigC*EPSc     ! GROWTH RATES FROM STATE: IS (=1 OR 2)

C--EXPONENT OF THE EXPONENTIAL FUNCTION IN THE 2D NORMAL DENSITY FUNCTION

      EXPONT1 = -(Gc-KaC(1))**2/(2*SigC**2)
      EXPONT2 = -(Gc-KaC(2))**2/(2*SigC**2)

      fy21 = DEXP(EXPONT2-EXPONT1)         ! ratio = fy(y,2)/fy(y,1)

C--BASIAN UPDATING: (13)

      BAYSIAN = (Q*MUt + (1-P)*fy21*(1-MUt))/(MUt+fy21*(1-MUt))    ! BAYESIAN UPDATE OF MU_{t+1}(y)

      RETURN
      END
C=================================================================================
C RETURN THE BAYSIAN UPDATED DENSITY, GIVEN DRIFT MU(1), MU(2) AND DIFFUSION SIGMA.
C INPUTS: Y---OBSERVED DIV AT $t+1$; PIT = Pi_t---Pr(S_{t+1}=1|F_t); MU---DIMENSION
C         OF 2, MU(1), MU(2), DRIFT OF DIV OF TWO STATES = {1, 2}; SIG---VOLATILITY;
C         P---TRANSITION PROB OF STATE 1 TO STATE 1; Q---TRAN. PROB OF STATE 2 TO 2;
C OUTPUTS: PITP1 = Pi_{t+1}---Pr(S_{t+2}=1|F_{t+1}), BAYESIAN UPDATED PROB GIVEM Y.
C=================================================================================
      SUBROUTINE DEN_RATIO_GRID(NQUD,QX,FY21)
      IMPLICIT NONE
      INTEGER NQUD,IS,IC
      REAL*8 QX(NQUD),FY21(2,NQUD)
      REAL*8 Zc,Zd,Gc,Gd,DET,C1,C2,D1,D2,EXPONT1,EXPONT2
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU

C--FACTOR SQRT(2) IS INTRODUCED TO CHANGE VARIABLE FROM NORMAL DENSITY FACTOR: EXP(X^2/),
C--TO A GUASIAN QUADRATURE FACTOR: EXP(-X^2). EXCEPT THE FACTOR SQRT(2), Zc AND Zd
C--ARE INDEPENDENT STD NORMAL. THE TRANSFORMATION IN Gc AND Gd YIELDS THE CORRECT
C--CORRELATION MATRIX. Zc AND Zd ARE Guassian quadrature points. INTEGRATION
C--(EXPECTATION) IS PERFORMED OVER THEM.

      DO IS = 1, 2
       DO IC = 1, NQUD
        Zc = QX(IC)
        Gc = KaC(IS) + SigC*DSQRT(2.0D0)*Zc     ! GROWTH RATES FROM STATE: IS (=1 OR 2)

C--EXPONENT OF THE EXPONENTIAL FUNCTION IN THE 2D NORMAL DENSITY FUNCTION
        EXPONT1 = -(Gc-KaC(1))**2/(2*SigC**2)
        EXPONT2 = -(Gc-KaC(2))**2/(2*SigC**2)

        FY21(IS,IC) = DEXP(EXPONT2-EXPONT1)         ! ratio = fy(y,2)/fy(y,1)
       ENDDO
      ENDDO

      RETURN
      END
C====================================================================
C GIVEN THE ORDER OF APPROXIMATION: NORD, AND THE COEFFICIENTS: CN(0:NORD),
C RETURN THE CHEBYSHEV POLYNORMIAL APPROXIMATION OF A FUNCTION.
C INPUTS: NORD---ORDER OF APPROXINATION;
C         CN(0:NORD)---APPROXIMATION COEF'S; INITIAL ORDER = 0; NORD <= 100;
C         THE REMAINING COEF'S FROM: (NORD+1):100, SET TO ZERO AND NOT USED.
C         Z---THE POINT AT WHICH THE VALUE OF THE FUNCTION IS APPROXIMATED;
C OUTPUTS: VALUE OF FUNCTION AT Z, WHICH IS [-1,1].
C====================================================================
      FUNCTION CBSVN(NORD,CN,Z)
      IMPLICIT NONE
      INTEGER NORD,I
      REAL*8 CBSVN,Z,C0,C1,C2,CN(0:100)

      IF (NORD.LT.2.OR.NORD.GT.100) STOP 'NEED 100 >= NORD >= 2'

      C0 = 1.0D0                       ! C0, C1: FIRST TWO CHEBYSHEV
      C1 = Z
      CBSVN = C0*CN(0) + C1*CN(1)      ! CBSVN = SUM_{I=0}^N T(I,Z)*CN(I)
      DO I = 2, NORD
       C2 = 2*Z*C1 - C0              ! RECURSIVELY DEFINED CHEBYSHEV T(N,Z)=2*Z*T(N-1,Z)-T(N-2,Z)
       CBSVN = CBSVN + C2*CN(I)      ! CHEBYSHEV APPROX WITH COEF COEF(0:N)
       C0 = C1                       ! UPDATE C0, C1 FOR NEXT ITERATION/LOOP
       C1 = C2
      ENDDO

      RETURN
      END
C========================================================================
C SOLVE THE FULL INFORMATION MODEL FOR THE VALUE FUNCTION; G(1) AND G(2).
C========================================================================
      SUBROUTINE FULL_REQS_G(NEQ,SN,RC)
      IMPLICIT NONE
      INTEGER NEQ
      REAL*8 SN(NEQ),RC(NEQ)
      REAL*8 EX1,EX2,EX3,H1,H2
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      EX1 = 1 - GAM
      EX2 = (1-RHO)/(1-GAM)
      EX3 = 1 - RHO

      H1 = SN(1)**EX1*DEXP(KaC(1)*EX1+0.5*EX1**2*SigC**2)
      H2 = SN(2)**EX1*DEXP(KaC(2)*EX1+0.5*EX1**2*SigC**2)

      RC(1) = (1-BET) + BET*(Q*H1+(1-Q)*H2)**EX2 - SN(1)**EX3
      RC(2) = (1-BET) + BET*((1-P)*H1+P*H2)**EX2 - SN(2)**EX3

      RETURN
      END
C========================================================================
C GIVEN THE VALUE FUNCTION: G(1) AND G(2), SOLVE THE FULL INFORMATION MODEL
C FOR THE PRICE DIV RATIO: varphi(1) AND varphi(2).
C========================================================================
      SUBROUTINE FULL_REQS_RM(G1,G2,VP1,VP2)
      IMPLICIT NONE
      REAL*8 G1,G2,VP1,VP2,COR
      REAL*8 EX1,EX2,EX3,VAR,CD1,CD2,C1,C2,DEN1,DEN2,X1,X2,Y1,Y2,DET
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 BET,GAM,RHO,ETA
      COMMON /ModelParam/BET,GAM,RHO,ETA

      EX1 = 1 - GAM
      EX2 = (RHO-GAM)/(1-GAM)
      EX3 = RHO - GAM

      VAR = GAM**2*SigC**2 + SigD**2 - 2*GAM*SigC*SigD
      CD1 = DEXP(-GAM*KaC(1)+KaD(1)+0.5*VAR)
      CD2 = DEXP(-GAM*KaC(2)+KaD(2)+0.5*VAR)
      C1 = DEXP(KaC(1)*EX1+0.5*EX1**2*SigC**2)
      C2 = DEXP(KaC(2)*EX1+0.5*EX1**2*SigC**2)
      DEN1 = (Q*G1**EX1*C1+(1-Q)*G2**EX1*C2)**EX2
      DEN2 = ((1-P)*G1**EX1*C1+P*G2**EX1*C2)**EX2

      X1 = BET*Q*G1**EX3*CD1*DEXP(0.5*SigU**2)/DEN1       ! PUT EXP(SigU^2/2) BACK
      Y1 = BET*(1-Q)*G2**EX3*CD2*DEXP(0.5*SigU**2)/DEN1
      X2 = BET*(1-P)*G1**EX3*CD1*DEXP(0.5*SigU**2)/DEN2
      Y2 = BET*P*G2**EX3*CD2*DEXP(0.5*SigU**2)/DEN2

      DET = (X1-1)*(Y2-1) - X2*Y1
      VP1 = -(Y2-Y1-1)/DET - 1
      VP2 = -(X1-X2-1)/DET - 1

      RETURN
      END
C=================================================================================
C RETURN N SERIES OF STATES, EACH WITH A SIZE OF M. MARKOV STATE MVS(N,M).
C P=PR(S1=1|S0=1), Q=PR(S1=2|S0=2); THEN PB=1-P=PR(S1=2|S0=1); QB=1-Q=PR(S1=1|S0=2)
C=================================================================================
      SUBROUTINE FULL_SIMURET(NR,BET,GAM,RHO,STATES,MNPI)
      IMPLICIT NONE
      INTEGER NR,STATES(NR),I,IS
      REAL*8 BET,GAM,RHO,MNPI(NR,8),RF,RE1,RE2,EM1,EM2

      DO I = 1, NR
       IS = STATES(I)
       CALL FULL_COND_MOMENTS(IS,BET,GAM,RHO,RF,RE1,RE2,EM1,EM2)

       MNPI(I,1) = RF                            ! COND. MEAN OF Rf ON MUt
       MNPI(I,2) = RF**2                         ! COND. SECOND MOMENT
       MNPI(I,3) = RE1                           ! COND. MEAN OF RE ON MUt
       MNPI(I,4) = RE2                           ! COND. SECOND MOMENT
       MNPI(I,5) = RE1 - RF                      ! COND. MEAN OF PRE (RE-RF) ON MUt
       MNPI(I,6) = RE2 + RF**2 - 2*RF*RE1        ! COND. SECOND OF (RE-RF)
       MNPI(I,7) = EM1                           ! COND. MEAN OF M_{t+1}
       MNPI(I,8) = EM2                           ! COND. SECOND MOMENT OF M_{t+1}
      ENDDO                ! END I LOOP

      RETURN
      END
C================================================================================================
C RETURN THE TWO FACTORS IN COMPUTING RISKFREE RATE AT EACH TIME POINT, GIVEN THE STATE PROB DIST.
C================================================================================================
      SUBROUTINE FULL_COND_MOMENTS(IS,BET,GAM,RHO,RF,RE1,RE2,EM1,EM2)
      IMPLICIT NONE
      INTEGER IS
      REAL*8 BET,GAM,RHO,RF,RE1,RE2,EM1,EM2,EM(2,2),RE(2,2)

C--COMPUTE COND. Rf AND MOMENTS OF PRICING KERNEL

      CALL FULL_M_MONTS(BET,GAM,RHO,EM,RE)

      RE1 = RE(IS,1)                        ! MEAN OF RM, COND. ON STATE: IS
      RE2 = RE(IS,2)                        ! MEAN OF RM^2

      EM1 = EM(IS,1)                        ! MEAN OF M_{t+1} COND. ON MUt
      EM2 = EM(IS,2)                        ! MEAN OF [M_{t+1}]^2
      RF = 1/EM1                            ! RISKFREE RATE

      RETURN
      END
C================================================================================================
C RETURN THE TWO FACTORS IN COMPUTING RISKFREE RATE AT EACH TIME POINT, GIVEN THE STATE PROB DIST.
C================================================================================================
      SUBROUTINE FULL_M_MONTS(BET,GAM,RHO,EM,RE)
      IMPLICIT NONE
      REAL*8 BET,GAM,RHO,EM(2,2),RE(2,2)
      REAL*8 D(2,2),X(2,2),C(2,2),EX1,EX2,EX3,K1,K2,DEN1,DEN2
      REAL*8 P,Q,KaC(2),KaD(2),SigC,SigD,SigU
      COMMON /DynamParam/P,Q,KaC,KaD,SigC,SigD,SigU
      REAL*8 G1,G2,VP1,VP2
      COMMON /FULL_G_RM_COEF/G1,G2,VP1,VP2

C--EXPECTATION OF DIV GROWTH RATES AND ITS SQUARE.
C--THE MATRIX D(I,J): FIRST INDEX: STATES; SECOND INDEX: MOMENTS (FIRST TWO)

      D(1,1) = DEXP(KaD(1)+0.5*SigD**2+0.5*SigU**2)    ! E_1[D_{t+1}/D_t], SUBSCRIPT DENOTES STATE AT: (t+1)
      D(2,1) = DEXP(KaD(2)+0.5*SigD**2+0.5*SigU**2)    ! E_2[D_{t+1}/D_t]
      D(1,2) = DEXP(2*KaD(1)+2*SigD**2+2*SigU**2)    ! E_1[(D_{t+1}/D_t)^2]
      D(2,2) = DEXP(2*KaD(2)+2*SigD**2+2*SigU**2)    ! E_2[(D_{t+1}/D_t)^2]

C--X(I,J) = (1+varphi(z_{t+1}))/varphi(z_t)

      X(1,1) = (VP1+1)/VP1
      X(1,2) = (VP2+1)/VP1
      X(2,1) = (VP1+1)/VP2
      X(2,2) = (VP2+1)/VP2

C--CONDITIONAL EXPECTATION OF EQUITY RETURN AND ITS SQUARE
C--THE MATRIX RE(I,J): FIRST INDEX: STATES; SECOND INDEX: MOMENTS (FIRST TWO)

      RE(1,1) = Q*X(1,1)*D(1,1) + (1-Q)*X(1,2)*D(2,1)        ! E_1[Rm], SUBSCRIPT DENOTES STATE AT: t
      RE(2,1) = (1-P)*X(2,1)*D(1,1) + P*X(2,2)*D(2,1)
      RE(1,2) = Q*X(1,1)**2*D(1,2) + (1-Q)*X(1,2)**2*D(2,2)  ! E_1[Rm^2]
      RE(2,2) = (1-P)*X(2,1)**2*D(1,2) + P*X(2,2)**2*D(2,2)

C--CONDITIONAL EXPECTATION OF (C_{t+1}/C_t)^{-gam} AND ITS SQUARE

      C(1,1) = DEXP(-GAM*KaC(1)+0.5*GAM**2*SigC**2)    ! E_1[(C_{t+1}/C_t)^{-gam}], SUBSCRIPT DENOTES STATE AT: (t+1)
      C(2,1) = DEXP(-GAM*KaC(2)+0.5*GAM**2*SigC**2)    ! E_2[(C_{t+1}/C_t)^{-gam}]
      C(1,2) = DEXP(-2*GAM*KaC(1)+2*GAM**2*SigC**2)    ! E_1[(C_{t+1}/C_t)^{-2gam}]
      C(2,2) = DEXP(-2*GAM*KaC(2)+2*GAM**2*SigC**2)    ! E_2[(C_{t+1}/C_t)^{-2gam}]

      EX1 = 1 - GAM
      EX2 = (RHO-GAM)/(1-GAM)
      EX3 = RHO - GAM

      K1 = DEXP(KaC(1)*EX1+0.5*EX1**2*SigC**2)      ! E_1[(C_{t+1}/C_t)^{1-gam}], SUBSCRIPT DENOTES STATE AT: (t+1)
      K2 = DEXP(KaC(2)*EX1+0.5*EX1**2*SigC**2)      ! E_2[(C_{t+1}/C_t)^{1-gam}]
      DEN1 = (Q*G1**EX1*K1+(1-Q)*G2**EX1*K2)**EX2    ! R_1(V_{t+1}), SUBSCRIPT DENOTES STATE AT: t
      DEN2 = ((1-P)*G1**EX1*K1+P*G2**EX1*K2)**EX2    ! R_2(V_{t+1})

C--CONDITIONAL EXPECTATION OF PRICING KERNEL

      EM(1,1) = BET*(Q*G1**EX3*C(1,1)+(1-Q)*G2**EX3*C(2,1))/DEN1   ! E_1[PRICING KERNEL], SUBSCRIPT DENOTES STATE AT: t
      EM(2,1) = BET*((1-P)*G1**EX3*C(1,1)+P*G2**EX3*C(2,1))/DEN2
      EM(1,2) = BET**2*(Q*G1**(2*EX3)*C(1,2) +
     &          (1-Q)*G2**(2*EX3)*C(2,2))/DEN1**2                  ! E_1[PRICING KERNEL SQUARED]
      EM(2,2) = BET**2*((1-p)*G1**(2*EX3)*C(1,2) +
     &          P*G2**(2*EX3)*C(2,2))/DEN2**2

      RETURN
      END
